@@ -16,8 +16,8 @@
 # permissions and limitations under the License.
 
 """
-Module ``dsl`` contains various functions to support the
-compilation of domain specific languages based on an EBNF-grammar.
+Module ``dsl`` contains high-level functions for the compilation
+of domain specific languages based on an EBNF-grammar.
 """
 
 import functools
@@ -33,20 +33,21 @@ from DHParser.ebnf import EBNFCompiler, grammar_changed, DHPARSER_IMPORTS, \
     get_ebnf_preprocessor, get_ebnf_grammar, get_ebnf_transformer, get_ebnf_compiler, \
     PreprocessorFactoryFunc, ParserFactoryFunc, TransformerFactoryFunc, CompilerFactoryFunc
 from DHParser.error import Error, is_error, has_errors, only_errors, \
-    CANNOT_VERIFY_TRANSTABLE_WARNING
+    CANNOT_VERIFY_TRANSTABLE_WARNING, ErrorCode, ERROR
 from DHParser.log import suspend_logging, resume_logging, is_logging, log_dir, append_log
 from DHParser.parse import Grammar
 from DHParser.preprocess import nil_preprocessor, PreprocessorFunc
 from DHParser.syntaxtree import Node
-from DHParser.transform import TransformationFunc, TransformationDict
+from DHParser.transform import TransformerCallable, TransformationDict
 from DHParser.toolkit import DHPARSER_DIR, DHPARSER_PARENTDIR, load_if_file, is_python_code, \
     compile_python_object, re, as_identifier, is_filename, relative_path
 from typing import Any, cast, List, Tuple, Union, Iterator, Iterable, Optional, \
-    Callable, Generator
+    Callable, Sequence
 
 
 __all__ = ('DefinitionError',
            'CompilationError',
+           'read_template',
            'load_compiler_suite',
            'compileDSL',
            'raw_compileEBNF',
@@ -54,9 +55,11 @@ __all__ = ('DefinitionError',
            'grammar_provider',
            'create_parser',
            'compile_on_disk',
-           'recompile_grammar')
+           'recompile_grammar',
+           'restore_server_script')
 
 
+@functools.lru_cache()
 def read_template(template_name: str) -> str:
     """
     Reads a script-template from a template file named `template_name`
@@ -84,14 +87,12 @@ AST_SECTION = "AST SECTION - Can be edited. Changes will be preserved."
 COMPILER_SECTION = "COMPILER SECTION - Can be edited. Changes will be preserved."
 END_SECTIONS_MARKER = "END OF DHPARSER-SECTIONS"
 
-DHPARSER_MAIN = read_template('DSLParser.pyi')
-
 
 class DSLException(Exception):
     """
     Base class for DSL-exceptions.
     """
-    def __init__(self, errors: Union[List[Error], Generator[Error, None, None]]):
+    def __init__(self, errors: Union[Sequence[Error], Iterator[Error]]):
         assert isinstance(errors, Iterator) or isinstance(errors, list) \
             or isinstance(errors, tuple)
         self.errors = list(errors)
@@ -127,6 +128,9 @@ class CompilationError(DSLException):
         self.AST = AST
         self.result = result
 
+    def __str__(self):
+        return '\n'.join(str(error) for error in self.errors)
+
 
 def error_str(messages: Iterable[Error]) -> str:
     """
@@ -150,12 +154,12 @@ def grammar_instance(grammar_representation) -> Tuple[Grammar, str]:
             parser_py = grammar_src  # type: str
             messages = []            # type: List[Error]
         else:
-            log_dir = suspend_logging()
+            lg_dir = suspend_logging()
             result, messages, _ = compile_source(
                 grammar_src, None,
                 get_ebnf_grammar(), get_ebnf_transformer(), get_ebnf_compiler())
             parser_py = cast(str, result)
-            resume_logging(log_dir)
+            resume_logging(lg_dir)
         if has_errors(messages):
             raise DefinitionError(only_errors(messages), grammar_src)
         imports = DHPARSER_IMPORTS.format(
@@ -179,8 +183,9 @@ def grammar_instance(grammar_representation) -> Tuple[Grammar, str]:
 def compileDSL(text_or_file: str,
                preprocessor: Optional[PreprocessorFunc],
                dsl_grammar: Union[str, Grammar],
-               ast_transformation: TransformationFunc,
-               compiler: Compiler) -> Any:
+               ast_transformation: TransformerCallable,
+               compiler: Compiler,
+               fail_when: ErrorCode = ERROR) -> Any:
     """
     Compiles a text in a domain specific language (DSL) with an
     EBNF-specified grammar. Returns the compiled text or raises a
@@ -195,13 +200,13 @@ def compileDSL(text_or_file: str,
     parser, grammar_src = grammar_instance(dsl_grammar)
     result, messages, AST = compile_source(text_or_file, preprocessor, parser,
                                            ast_transformation, compiler)
-    if has_errors(messages):
+    if has_errors(messages, fail_when):
         src = load_if_file(text_or_file)
-        raise CompilationError(only_errors(messages), src, grammar_src, AST, result)
+        raise CompilationError(only_errors(messages, fail_when), src, grammar_src, AST, result)
     return result
 
 
-def raw_compileEBNF(ebnf_src: str, branding="DSL") -> EBNFCompiler:
+def raw_compileEBNF(ebnf_src: str, branding="DSL", fail_when: ErrorCode = ERROR) -> EBNFCompiler:
     """
     Compiles an EBNF grammar file and returns the compiler object
     that was used and which can now be queried for the result as well
@@ -220,7 +225,7 @@ def raw_compileEBNF(ebnf_src: str, branding="DSL") -> EBNFCompiler:
     grammar = get_ebnf_grammar()
     compiler = get_ebnf_compiler(branding, ebnf_src)
     transformer = get_ebnf_transformer()
-    compileDSL(ebnf_src, nil_preprocessor, grammar, transformer, compiler)
+    compileDSL(ebnf_src, nil_preprocessor, grammar, transformer, compiler, fail_when)
     return compiler
 
 
@@ -247,17 +252,21 @@ def compileEBNF(ebnf_src: str, branding="DSL") -> str:
     compiler = raw_compileEBNF(ebnf_src, branding)
     src = ["#/usr/bin/python\n",
            SECTION_MARKER.format(marker=SYMBOLS_SECTION),
-           DHPARSER_IMPORTS.format(dhparser_parentdir = relative_dhpath),
+           DHPARSER_IMPORTS.format(dhparser_parentdir=relative_dhpath),
            SECTION_MARKER.format(marker=PREPROCESSOR_SECTION), compiler.gen_preprocessor_skeleton(),
            SECTION_MARKER.format(marker=PARSER_SECTION), compiler.result,
            SECTION_MARKER.format(marker=AST_SECTION), compiler.gen_transformer_skeleton(),
            SECTION_MARKER.format(marker=COMPILER_SECTION), compiler.gen_compiler_skeleton(),
-           SECTION_MARKER.format(marker=SYMBOLS_SECTION), DHPARSER_MAIN.format(NAME=branding)]
+           SECTION_MARKER.format(marker=END_SECTIONS_MARKER),
+           read_template('DSLParser.pyi').format(NAME=branding)]
     return '\n'.join(src)
 
 
 @functools.lru_cache()
-def grammar_provider(ebnf_src: str, branding="DSL", additional_code: str='') -> ParserFactoryFunc:
+def grammar_provider(ebnf_src: str,
+                     branding="DSL",
+                     additional_code: str = '',
+                     fail_when: ErrorCode = ERROR) -> ParserFactoryFunc:
     """
     Compiles an EBNF-grammar and returns a grammar-parser provider
     function for that grammar.
@@ -275,15 +284,13 @@ def grammar_provider(ebnf_src: str, branding="DSL", additional_code: str='') ->
         A provider function for a grammar object for texts in the
         language defined by ``ebnf_src``.
     """
-    grammar_src = compileDSL(ebnf_src, nil_preprocessor, get_ebnf_grammar(),
-                             get_ebnf_transformer(), get_ebnf_compiler(branding, ebnf_src))
+    grammar_src = compileDSL(
+        ebnf_src, nil_preprocessor, get_ebnf_grammar(), get_ebnf_transformer(),
+        get_ebnf_compiler(branding, ebnf_src), fail_when)
     log_name = get_config_value('compiled_EBNF_log')
-    if log_name:
-        if is_logging():
+    if log_name and is_logging():
             append_log(log_name, grammar_src)
-        else:
-            print(grammar_src)
-    imports = DHPARSER_IMPORTS.format(dhparser_parentdir = relative_path('.', DHPARSER_PARENTDIR))
+    imports = DHPARSER_IMPORTS.format(dhparser_parentdir=relative_path('.', DHPARSER_PARENTDIR))
     grammar_factory = compile_python_object('\n'.join([imports, additional_code, grammar_src]),
                                             r'get_(?:\w+_)?grammar$')
     if callable(grammar_factory):
@@ -292,15 +299,17 @@ def grammar_provider(ebnf_src: str, branding="DSL", additional_code: str='') ->
     raise ValueError('Could not compile grammar provider!')
 
 
-def create_parser(ebnf_src: str, branding="DSL", additional_code: str='') -> Grammar:
+def create_parser(ebnf_src: str, branding="DSL", additional_code: str = '') -> Grammar:
     """Compiles the ebnf source into a callable Grammar-object. This is
     essentially syntactic sugar for `grammar_provider(ebnf)()`.
     """
     grammar_factory = grammar_provider(ebnf_src, branding, additional_code)
-    return grammar_factory()
+    grammar = grammar_factory()
+    grammar.python_src__ = grammar_factory.python_src__
+    return grammar
 
 
-def split_source(file_name: str, file_content: str) -> Tuple[str]:
+def split_source(file_name: str, file_content: str) -> List[str]:
     """Splits the `file_content` into the seven sections: intro, imports,
     preprocessor_py, parser_py, ast_py, compiler_py, outro.
     Raises a value error, if the number of sections if not equal to 7.
@@ -328,7 +337,7 @@ def load_compiler_suite(compiler_suite: str) -> \
     assert isinstance(compiler_suite, str)
     source = load_if_file(compiler_suite)
     dhpath = relative_path(os.path.dirname('.'), DHPARSER_PARENTDIR)
-    imports = DHPARSER_IMPORTS.format(dhparser_parentdir = dhpath)
+    imports = DHPARSER_IMPORTS.format(dhparser_parentdir=dhpath)
     if is_python_code(compiler_suite):
         sections = split_source(compiler_suite, source)
         _, imports, preprocessor_py, parser_py, ast_py, compiler_py, _ = sections
@@ -340,11 +349,11 @@ def load_compiler_suite(compiler_suite: str) -> \
     else:
         # Assume source is an ebnf grammar.
         # Is there really any reasonable application case for this?
-        log_dir = suspend_logging()
+        lg_dir = suspend_logging()
         compiler_py, messages, _ = compile_source(source, None, get_ebnf_grammar(),
                                                   get_ebnf_transformer(),
                                                   get_ebnf_compiler(compiler_suite, source))
-        resume_logging(log_dir)
+        resume_logging(lg_dir)
         if has_errors(messages):
             raise DefinitionError(only_errors(messages), source)
         preprocessor = get_ebnf_preprocessor
@@ -360,7 +369,7 @@ def is_outdated(compiler_suite: str, grammar_source: str) -> bool:
     """
     Returns ``True``  if the ``compile_suite`` needs to be updated.
 
-    An update is needed, if either the grammar in the compieler suite
+    An update is needed, if either the grammar in the compiler suite
     does not reflect the latest changes of ``grammar_source`` or if
     sections from the compiler suite have diligently been overwritten
     with whitespace order to trigger their recreation. Note: Do not
@@ -382,7 +391,7 @@ def is_outdated(compiler_suite: str, grammar_source: str) -> bool:
         return True
 
 
-def run_compiler(text_or_file: str, compiler_suite: str) -> Any:
+def run_compiler(text_or_file: str, compiler_suite: str, fail_when: ErrorCode = ERROR) -> Any:
     """Compiles a source with a given compiler suite.
 
     Args:
@@ -402,7 +411,7 @@ def run_compiler(text_or_file: str, compiler_suite: str) -> Any:
         CompilerError
     """
     preprocessor, parser, ast, compiler = load_compiler_suite(compiler_suite)
-    return compileDSL(text_or_file, preprocessor(), parser(), ast(), compiler())
+    return compileDSL(text_or_file, preprocessor(), parser(), ast(), compiler(), fail_when)
 
 
 def compile_on_disk(source_file: str, compiler_suite="", extension=".xml") -> Iterable[Error]:
@@ -468,8 +477,7 @@ def compile_on_disk(source_file: str, compiler_suite="", extension=".xml") -> It
         # either an EBNF- or no compiler suite given
         ebnf_compiler = cast(EBNFCompiler, compiler1)  # type: EBNFCompiler
         global SECTION_MARKER, RX_SECTION_MARKER, PREPROCESSOR_SECTION, PARSER_SECTION, \
-            AST_SECTION, COMPILER_SECTION, END_SECTIONS_MARKER, RX_WHITESPACE, \
-            DHPARSER_MAIN
+            AST_SECTION, COMPILER_SECTION, END_SECTIONS_MARKER, RX_WHITESPACE
         f = None
         try:
             parser_name = rootname + 'Parser.py'
@@ -488,7 +496,7 @@ def compile_on_disk(source_file: str, compiler_suite="", extension=".xml") -> It
                               'Possibly due to a forgotten import at the beginning ' \
                               'of the AST-Block (!)'.format(str(e))
                 else:
-                    err_str = 'Exception {} while compiling AST-Tansformation: {}' \
+                    err_str = 'Exception {} while compiling AST-Transformation: {}' \
                               .format(str(type(e)), str(e))
                 messages.append(Error(err_str, 0, CANNOT_VERIFY_TRANSTABLE_WARNING))
                 if is_logging():
@@ -507,7 +515,7 @@ def compile_on_disk(source_file: str, compiler_suite="", extension=".xml") -> It
         if RX_WHITESPACE.fullmatch(intro):
             intro = '#!/usr/bin/env python3'
         if RX_WHITESPACE.fullmatch(outro):
-            outro = DHPARSER_MAIN.format(NAME=compiler_name)
+            outro = read_template('DSLParser.pyi').format(NAME=compiler_name)
         if RX_WHITESPACE.fullmatch(imports):
             imports = DHParser.ebnf.DHPARSER_IMPORTS.format(dhparser_parentdir=dhpath)
         if RX_WHITESPACE.fullmatch(preprocessor):
@@ -619,3 +627,35 @@ def recompile_grammar(ebnf_filename, force=False,
     if not messages and os.path.exists(error_file_name):
         os.remove(error_file_name)
     return True
+
+
+def restore_server_script(ebnf_filename: str, overwrite: bool = False):
+    """Creates a script for compiling texts adhering to the given grammar
+    with a server. Because the server script relies on a parser script,
+    a parser scripte will be created, too, if it does not yet exist.
+
+    :var ebnf_filename: The filename of the grammar, from which the servfer
+        script's filename is derived.
+    :var overwrite: If True an existing server script will be overwritten.
+    """
+    if os.path.isdir(ebnf_filename):
+        for entry in os.listdir(ebnf_filename):
+            if entry.lower().endswith('.ebnf') and os.path.isfile(entry):
+                restore_server_script(entry)
+        return
+
+    base, _ = os.path.splitext(ebnf_filename)
+    name = os.path.basename(base)
+    server_name = base + 'Server.py'
+    compiler_name = base + 'Parser.py'
+    if not os.path.exists(server_name) or overwrite:
+        template = read_template('DSLServer.pyi')
+        reldhparserdir = os.path.relpath(os.path.dirname(DHPARSER_DIR), os.path.abspath('.'))
+        serverscript = base + 'Server.py'
+        with open(serverscript, 'w') as f:
+            f.write(template.replace('DSL', name).replace('RELDHPARSERDIR', reldhparserdir))
+        if platform.system() != "Windows":
+            # set file permissions so that the compilerscript can be executed
+            st = os.stat(serverscript)
+            os.chmod(serverscript, st.st_mode | stat.S_IEXEC)
+    if not os.path.exists(compiler_name):  recompile_grammar(ebnf_filename)
