@@ -37,7 +37,7 @@ to which either ``False`` (turn off logging), a log directory name or
 The other components of DHParser check whether logging is on and
 write log files in the the logging directory accordingly. Usually,
 this will be concrete and abstract syntax trees as well as the full
-and abreviated parsing history.
+and abbreviated parsing history.
 
 Example::
 
@@ -53,16 +53,17 @@ Example::
 
 import collections
 import contextlib
+import copy
 import html
 import os
 from typing import List, Tuple, Union, Optional
 
 from DHParser.configuration import access_presets, finalize_presets, get_config_value, \
-    set_config_value
+    set_config_value, get_preset_value, set_preset_value
 from DHParser.error import Error
 from DHParser.stringview import StringView
 from DHParser.syntaxtree import Node, FrozenNode, ZOMBIE_TAG, EMPTY_PTYPE
-from DHParser.toolkit import escape_control_characters, abbreviate_middle
+from DHParser.toolkit import escape_ctrl_chars, abbreviate_middle, cython
 
 __all__ = ('CallItem',
            'start_logging',
@@ -76,7 +77,6 @@ __all__ = ('CallItem',
            'clear_logs',
            'NONE_TAG',
            'NONE_NODE',
-           'freeze_callstack',
            'HistoryRecord',
            'log_ST',
            'log_parsing_history')
@@ -95,12 +95,12 @@ def start_logging(dirname: str = "LOGS"):
     """Turns logging on an sets the log-directory to `dirname`.
     The log-directory, if it does not already exist, will be created
     lazily, i.e. only when logging actually starts."""
-    CFG = access_presets()
+    access_presets()
     log_dir = os.path.abspath(dirname) if dirname else ''
-    if log_dir != CFG['log_dir']:
-        CFG['log_dir'] = log_dir
+    if log_dir != get_preset_value('log_dir'):
+        set_preset_value('log_dir', log_dir)
         set_config_value('log_dir', log_dir)
-        finalize_presets()
+    finalize_presets()
 
 
 def suspend_logging() -> str:
@@ -114,8 +114,9 @@ def suspend_logging() -> str:
 def resume_logging(log_dir: str = ''):
     """Resumes logging in the current thread with the given log-dir."""
     if not 'log_dir':
-        CFG = access_presets()
-        log_dir = CFG['log_dir']
+        access_presets()
+        log_dir = get_preset_value('log_dir')
+        finalize_presets()
     set_config_value('log_dir', log_dir)
 
 
@@ -179,14 +180,16 @@ def is_logging(thread_local_query: bool = True) -> bool:
     if thread_local_query:
         return bool(get_config_value('log_dir'))
     else:
-        CFG = access_presets()
-        return bool(CFG['log_dir'])
+        access_presets()
+        result = bool(get_preset_value('log_dir'))
+        finalize_presets()
+        return result
 
 
 def create_log(log_name: str) -> str:
     """
     Creates a new log file. If log_name is not just a file name but a path with
-    at least one directoy (which can be './') the file is not created in the
+    at least one directory (which can be './') the file is not created in the
     configured log directory but at the given path. If a file with the same
     name already exists, it will be overwritten.
 
@@ -199,6 +202,8 @@ def create_log(log_name: str) -> str:
     if not ldir:
         ldir = log_dir()
     if ldir:
+        if not os.path.exists(ldir):
+            os.makedirs(ldir)
         with open(os.path.join(ldir, file_name), 'w', encoding='utf-8') as f:
             f.write('LOG-FILE: ' + log_name + '\n\n')
         return log_name
@@ -216,10 +221,10 @@ def append_log(log_name: str, *strings, echo: bool = False) -> None:
         No delimiters will be added, i.e. all delimiters like blanks or
         linefeeds need to be added explicitly to the list of strings, before
         calling `append_log()`.
-    :param echo: If True, the log message will be echoed on the terminal. This
-        will also happen if logging is turned off.
+    :param echo: If True, the log message will be echoed on the terminal.
+        In this case the message will even be echoed if logging is turned off.
     """
-    ldir, _ = os.path.split(log_name)
+    ldir, log_name = os.path.split(log_name)
     if not ldir:
         ldir = log_dir()
     if ldir and log_name:
@@ -229,6 +234,7 @@ def append_log(log_name: str, *strings, echo: bool = False) -> None:
         with open(log_path, 'a', encoding='utf-8') as f:
             for text in strings:
                 f.write(text)
+            f.flush()
     if echo:
         print(''.join(strings))
 
@@ -258,14 +264,13 @@ def clear_logs(logfile_types=frozenset(['.cst', '.ast', '.log'])):
 #
 #######################################################################
 
+class NoneNode(FrozenNode):
+    def __bool__(self):
+        return False
+
 
 NONE_TAG = ":None"
-NONE_NODE = FrozenNode(NONE_TAG, '')
-
-
-def freeze_callstack(call_stack: List[CallItem]) -> Tuple[CallItem, ...]:
-    """Returns a frozen copy of the call stack."""
-    return tuple((tn, pos) for tn, pos in call_stack if tn != ":Forward")
+NONE_NODE = NoneNode(NONE_TAG, '')
 
 
 class HistoryRecord:
@@ -319,12 +324,7 @@ class HistoryRecord:
                  text: StringView,
                  line_col: Tuple[int, int],
                  errors: List[Error] = []) -> None:
-        # copy call stack, dropping uninformative Forward-Parsers
-        # self.call_stack = call_stack    # type: Tuple[CallItem,...]
-        if isinstance(call_stack, tuple):
-            self.call_stack = call_stack  # type: Tuple[CallItem,...]
-        else:
-            self.call_stack = freeze_callstack(call_stack)
+        self.call_stack = tuple(call_stack)              # type: Tuple[CallItem, ...]
         self.node = NONE_NODE if node is None else node  # type: Node
         self.text = text                                 # type: StringView
         self.line_col = line_col                         # type: Tuple[int, int]
@@ -352,6 +352,7 @@ class HistoryRecord:
         """
         return '"{}", "{}", "{}", "{}", "{}"'.format(*self.as_tuple())
 
+    @cython.locals(i=cython.int, k=cython.int, n=cython.int)
     def as_html_tr(self) -> str:
         """
         Returns history record formatted as an html table row.
@@ -360,21 +361,21 @@ class HistoryRecord:
             '-&gt;', '<span>&shy;-&gt;</span>')
         status = html.escape(self.status)
         excerpt = html.escape(self.excerpt)
-        if excerpt.startswith(' '):
+        if excerpt[:1] == ' ':
             stripped = excerpt.lstrip()
             excerpt = "&nbsp;" * (len(excerpt) - len(stripped)) + stripped
         classes = list(HistoryRecord.Snapshot_Fields)
         idx = {field_name: i for i, field_name in enumerate(classes)}
-        classes[idx['status']] = 'error' if status.startswith('ERROR') else status.lower()
+        classes[idx['status']] = 'error' if status[:5] == 'ERROR' else status.lower()
         if status in (self.MATCH, self.DROP):
             n = max(40 - len(excerpt), 0)
             dots = '...' if len(self.text) > n else ''
             excerpt = excerpt + '<span class="unmatched">' + self.text[:n] + dots + '</span>'
             i = stack.rfind('-&gt;')
-            chr = stack[i + 12:i + 13]
-            while not chr.isidentifier() and i >= 0:
+            ch = stack[i + 12:i + 13]
+            while not ch.isidentifier() and i >= 0:
                 i = stack.rfind('-&gt;', 0, i)
-                chr = stack[i + 12:i + 13]
+                ch = stack[i + 12:i + 13]
             if i >= 0:
                 i += 12
                 k = stack.find('<', i)
@@ -404,8 +405,8 @@ class HistoryRecord:
         short_stack = []
         anonymous_tail = True
         for tag_name, _ in reversed(self.call_stack):
-            if tag_name.startswith(':'):
-                if anonymous_tail:
+            if tag_name[:1] == ':':
+                if anonymous_tail and tag_name != ":Forward":
                     short_stack.append(tag_name)
             else:
                 short_stack.append(tag_name)
@@ -432,7 +433,7 @@ class HistoryRecord:
         else:
             s = self.text
             excerpt = s[:36] + ' ...' if len(s) > 36 else s
-        excerpt = escape_control_characters(str(excerpt))
+        excerpt = escape_ctrl_chars(str(excerpt))
         return excerpt
 
     # @property
@@ -481,6 +482,27 @@ class HistoryRecord:
         return result
 
 
+    @staticmethod
+    def most_advanced_fail(history: List['HistoryRecord']) -> Union['HistoryRecord', None]:
+        """
+        Returns the closest-to-the-end-fail from the parsing-history.
+        Args:
+            history:  the parsing-history as a list of HistoryRecord objects
+
+        Returns:
+            the history record of the closest-to-the-end-fail or none if either history is
+            empty or no parser failed
+        """
+        remaining = -1
+        result = None
+        for record in history:
+            if (record.status == HistoryRecord.FAIL
+                    and (record.remaining < remaining or remaining < 0)):
+                result = record
+                remaining = record.remaining
+        return result
+
+
 #######################################################################
 #
 #  context specific log functions, i.e. logging of syntax trees,
@@ -506,10 +528,6 @@ def log_ST(syntax_tree, log_file_name) -> bool:
     return False
 
 
-LOG_SIZE_THRESHOLD = 10000    # maximum number of history records to log
-LOG_TAIL_THRESHOLD = 500      # maximum number of history recors for "tail log"
-
-
 def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) -> bool:
     """
     Writes a log of the parsing history of the most recently parsed document, if
@@ -558,24 +576,21 @@ def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) ->
     elif log_file_name.lower().endswith('.log'):
         log_file_name = log_file_name[:-4]
 
-    full_history = ['<h1>Full parsing history of "%s"</h1>' % log_file_name]  # type: List[str]
-
+    history = ['<h1>Parsing history of "%s"</h1>' % log_file_name]  # type: List[str]
+    LOG_SIZE_THRESHOLD = get_config_value('log_size_threshold')
     if len(grammar.history__) > LOG_SIZE_THRESHOLD:
         warning = ('Sorry, man, %iK history records is just too many! '
                    'Only looking at the last %iK records.'
                    % (len(grammar.history__) // 1000, LOG_SIZE_THRESHOLD // 1000))
         html_warning = '<p><strong>' + warning + '</strong></p>'
-        full_history.append(html_warning)
+        history.append(html_warning)
 
     lead_in = '\n'. join(['<table>', HistoryRecord.COLGROUP, HistoryRecord.HEADINGS])
-    full_history.append(lead_in)
+    history.append(lead_in)
 
     for record in grammar.history__[-LOG_SIZE_THRESHOLD:]:
         line = record.as_html_tr() if html else (str(record) + '\n')
-        append_line(full_history, line)
+        append_line(history, line)
 
-    write_log(full_history, log_file_name + '_full')
-    if len(full_history) > LOG_TAIL_THRESHOLD + 10:
-        heading = '<h1>Last 500 records of parsing history of "%s"</h1>' % log_file_name + lead_in
-        write_log([heading] + full_history[-LOG_TAIL_THRESHOLD:], log_file_name + '_full.tail')
+    write_log(history, log_file_name)
     return True
