@@ -26,16 +26,17 @@ from multiprocessing import Pool
 
 scriptpath = os.path.dirname(__file__) or '.'
 sys.path.append(os.path.abspath(os.path.join(scriptpath, '..')))
+scriptpath = os.path.abspath(scriptpath)
 
-from DHParser.toolkit import compile_python_object, re, DHPARSER_PARENTDIR
+from DHParser.toolkit import compile_python_object, re, DHPARSER_PARENTDIR, \
+    normalize_circular_path
 from DHParser.preprocess import nil_preprocessor
 from DHParser import compile_source, INFINITE, Interleave
-from DHParser.configuration import access_thread_locals, get_config_value, \
-    EBNF_ANY_SYNTAX_HEURISTICAL, EBNF_ANY_SYNTAX_STRICT, EBNF_CLASSIC_SYNTAX, \
-    EBNF_REGULAR_EXPRESSION_SYNTAX, EBNF_PARSING_EXPRESSION_GRAMMAR_SYNTAX, set_config_value
+from DHParser.configuration import get_config_value, set_config_value
 from DHParser.error import has_errors, MANDATORY_CONTINUATION, PARSER_STOPPED_BEFORE_END, \
     REDEFINED_DIRECTIVE, UNUSED_ERROR_HANDLING_WARNING, AMBIGUOUS_ERROR_HANDLING, \
-    REORDERING_OF_ALTERNATIVES_REQUIRED, BAD_ORDER_OF_ALTERNATIVES
+    REORDERING_OF_ALTERNATIVES_REQUIRED, BAD_ORDER_OF_ALTERNATIVES, UNCONNECTED_SYMBOL_WARNING, \
+    ERROR, WARNING
 from DHParser.syntaxtree import WHITESPACE_PTYPE
 from DHParser.ebnf import get_ebnf_grammar, get_ebnf_transformer, EBNFTransform, \
     EBNFDirectives, get_ebnf_compiler, compile_ebnf, DHPARSER_IMPORTS, parse_ebnf, transform_ebnf
@@ -80,6 +81,20 @@ class TestDirectives:
         syntax_tree = parser("3 + 4 \n\n * 12")
         assert syntax_tree.errors_sorted
 
+    def test_whitespace_comments(self):
+        lang = "@whitespace = linefeed\n@comment = /%.*/\n" + self.mini_language
+        MinilangParser = grammar_provider(lang)
+        parser = MinilangParser()
+        assert parser
+        syntax_tree = parser("3 + 4 \n \n * 12")
+        assert syntax_tree.errors_sorted
+        syntax_tree = parser("3 + 4 \n  %comment \n * 12")
+        assert not syntax_tree.errors_sorted
+        syntax_tree = parser("3 + 4 \n  %comment\n\n * 12")
+        assert syntax_tree.errors_sorted
+        syntax_tree = parser("3 + 4 \n\n%comment\n * 12")
+        assert syntax_tree.errors_sorted
+
     def test_whitespace_vertical(self):
         lang = "@ whitespace = vertical\n" + self.mini_language
         parser = grammar_provider(lang)()
@@ -116,6 +131,18 @@ class TestDirectives:
         st = parser('')
         assert not st.errors and str(st) == ''
 
+    def test_drop(self):
+        lang = r"""
+            @ drop = backticked, whitespace
+            @ literalws = right
+            doc  = "*" word `*`
+            word = /\w+/
+        """
+        parser = create_parser(lang)
+        st = parser('* Hund*')
+        assert str(st) == "*Hund"
+
+
 
 class TestReservedSymbols:
     def test_comment_usage(self):
@@ -262,13 +289,38 @@ class TestCompilerErrors:
         assert messages
 
     def test_no_error(self):
-        """But reserved symbols should not be repoted as undefined.
+        """But reserved symbols should not be reported as undefined.
         """
         ebnf = """nothing =  WSP_RE__ | COMMENT__\n"""
-        result, messages, st = compile_source(ebnf, None, get_ebnf_grammar(),
-            get_ebnf_transformer(), get_ebnf_compiler('UndefinedSymbols'))
+        result, messages, st = compile_source(
+            ebnf, None, get_ebnf_grammar(), get_ebnf_transformer(),
+            get_ebnf_compiler('UndefinedSymbols'))
         assert not bool(messages), messages
 
+    def test_unconnected_symbols(self):
+        ebnf = """
+            root = "ROOT" | connected
+            connected = "CONNECTED"
+            unconnected = "UNCONNECTED (warning expected)"
+        """
+        result, messages, st = compile_source(
+            ebnf, None, get_ebnf_grammar(), get_ebnf_transformer(),
+            get_ebnf_compiler('UndefinedSymbols'), preserve_AST=False)
+        assert len(messages) == 1
+        assert messages[0].code == UNCONNECTED_SYMBOL_WARNING
+
+    def test_unconnected_symbols_corrupted_source(self):
+        ebnf = """
+            noroot
+            root = "ROOT" | connected
+            connected = "CONNECTED"
+            unconnected = "UNCONNECTED (warning expected)"
+        """
+        result, messages, st = compile_source(
+            ebnf, None, get_ebnf_grammar(), get_ebnf_transformer(),
+            get_ebnf_compiler('UndefinedSymbols'), preserve_AST=True)
+        assert any(m.code >= ERROR for m in messages)
+
 
 EBNF = r"""
 # EBNF-Grammar in EBNF
@@ -277,7 +329,7 @@ EBNF = r"""
 @ whitespace = /\s*/                            # whitespace includes linefeed
 @ literalws  = right                            # trailing whitespace of literals will be ignored tacitly
 @ drop       = whitespace                       # do not include whitespace in concrete syntax tree
-@ anonymous  = pure_elem, element
+@ disposable = pure_elem, element
 
 #: top-level
 
@@ -340,8 +392,8 @@ class TestSelfHosting:
         compiler_name = "EBNF"
         compiler = get_ebnf_compiler(compiler_name, EBNF)
         parser = get_ebnf_grammar()
-        result, errors, syntax_tree = compile_source(EBNF, None, parser,
-                                            get_ebnf_transformer(), compiler)
+        result, errors, syntax_tree = compile_source(
+            EBNF, None, parser, get_ebnf_transformer(), compiler)
         assert not errors, str(errors)
         # compile the grammar again using the result of the previous
         # compilation as parser
@@ -410,9 +462,9 @@ class TestBoundaryCases:
             assert False, "Grammar objects should be able to cope with unconnected parsers!"
         try:
             nonexistant = grammar['nonexistant']
-            assert False, "Grammar object shoul raise a KeyError if subscripted by " \
+            assert False, "Grammar object should raise an AttributeError if subscripted by " \
                           "a non-existant parser name!"
-        except KeyError:
+        except AttributeError:
             pass
 
 
@@ -427,16 +479,16 @@ class TestSynonymDetection:
         assert grammar('b').as_sxpr() == '(a (b "b"))'
 
     def test_synonym_anonymous_elimination(self):
-        ebnf = r"""@ anonymous = /_\w+$/
+        ebnf = r"""@ disposable = /_\w+$/
                   a = _b
                   _b = /b/
         """
         grammar = grammar_provider(ebnf)()
-        assert not grammar['a'].anonymous
-        assert grammar['_b'].anonymous
+        assert not grammar['a'].disposable
+        assert grammar['_b'].disposable
         assert grammar['a'].pname == 'a', grammar['a'].pname
         assert grammar['_b'].pname == '_b', grammar['_b'].pname
-        assert grammar('b').as_sxpr() == '(a "b")'
+        assert grammar('b').as_sxpr() == '(a "b")', grammar('b').as_sxpr()
 
 
 class TestFlowControlOperators:
@@ -521,9 +573,6 @@ class TestInterleave:
                'form_4 = 2*"d"' \
                'form_5 = ["e"]*3' \
                'form_6 = 5*{"f"}+'
-        # ast = parse_ebnf(ebnf)
-        # transform_ebnf(ast)
-        # print(ast.as_sxpr())
         grammar = create_parser(ebnf)
         assert grammar.form_6.repetitions == (5, INFINITE)
         assert grammar.form_5.repetitions == (0, 3)
@@ -561,6 +610,8 @@ class TestInterleave:
         assert len(grammar.prefix.parsers) > 1
         assert grammar('B A').content == 'B A'
         assert grammar('B').content == 'B'
+        result = grammar('')
+        assert result.content == '' and not result.errors
 
     def test_interleave_counted(self):
         ebnf = 'prefix = "A"{1,5} ° "B"{2,3}'
@@ -662,9 +713,9 @@ class TestErrorCustomization:
 class TestErrorCustomizationErrors:
     def test_ambiguous_error_customization(self):
         lang = """
-            document = series 
+            document = series
             @series_error = "ambiguous error message: does it apply to first or second '§'?"
-            series = "A" § "B" "C" | "X" § "Y" "Z" 
+            series = "A" § "B" "C" | "X" § "Y" "Z"
             """
         try:
             parser = grammar_provider(lang)()
@@ -687,32 +738,32 @@ class TestErrorCustomizationErrors:
         lang = r"""
             @comment = /(?:\/\/.*)|(?:\/\*(?:.|\n)*?\*\/)/
             document = alles
-            
+
             @alles_skip = /E|3|$/
             alles = (anfang | "1") (mitte | "2") (ende | "3") [COMMENT__]
-            
+
             @anfang_resume = /M|2/
             anfang = `A` §"NFANG"
-            
+
             @mitte_resume = /E|3/
             mitte = `M` §"ITTE"
-            
+
             ende = "ENDE"
         """
         assert lang.find("@mitte_") >= 0
         result, messages, ast = compile_ebnf(lang)
-        assert any(m.code == UNUSED_ERROR_HANDLING_WARNING for m in messages)
+        assert any(m.code == UNUSED_ERROR_HANDLING_WARNING for m in messages), str(messages)
 
         l2 = [zeile for zeile in lang.split('\n') if not zeile.lstrip().startswith('@mitte_')]
         lang2 = '\n'.join(l2)
         assert lang2.find('@mitte_') < 0
         result, messages, ast = compile_ebnf(lang2)
-        assert not messages
+        assert not messages, str(messages)
 
         l3 = [zeile for zeile in l2 if not zeile.lstrip().startswith('mitte')]
         lang3 = '\n'.join(l3).replace('mitte', '(`M` §"ITTE")')
         result, messages, ast = compile_ebnf(lang3)
-        assert not messages
+        assert not messages, str(messages)
 
         parser = create_parser(lang3)
         cst = parser('ANFANGMITTEENDE')
@@ -721,7 +772,6 @@ class TestErrorCustomizationErrors:
         assert cst.errors
         assert 'alles' in cst and 'ZOMBIE__' in cst['alles'] and 'ende' in cst['alles']
 
-
     def test_multiple_resume_definitions(self):
         lang = """
             document = series
@@ -761,9 +811,10 @@ class TestErrorCustomizationErrors:
                             'several strings'
             series = "A" § "B" "C"
             """
-        parser = grammar_provider(lang)()
+        provider = grammar_provider(lang)
+        parser = provider()
         result = parser('ADX')
-        assert "several strings" in str(result.errors)
+        assert "several strings" in str(result.errors), str(result.errors)
 
 
 class TestCustomizedResumeParsing:
@@ -782,10 +833,9 @@ class TestCustomizedResumeParsing:
             cba = "c" "b" §"a"
         GAMMA_RE = /(?=GA\w+)/
         """
-    gr = grammar_provider(lang)()
 
     def test_several_resume_rules_innermost_rule_matching(self):
-        gr = self.gr
+        gr = grammar_provider(self.lang, fail_when=WARNING)()
         content = 'ALPHA abc BETA bad GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
@@ -837,6 +887,50 @@ class TestCustomizedResumeParsing:
         st = grammar(doc3)
         assert st.children and st.children[-1].tag_name == 'word'
 
+class TestCustomizedResumeParsing_with_Parsers:
+    lang = r"""@ literalws = right
+        @ alpha_resume = ALPHA_RESUME
+        @ beta_resume = GAMMA_RE
+        @ bac_resume = /(?=GA\w+)/
+        document = alpha [beta] gamma "."
+        alpha = "ALPHA" abc
+            abc = §"a" "b" "c"
+          beta = "BETA" (bac | bca)
+            bac = "b" "a" §"c"
+            bca = "b" "c" §"a"
+          gamma = "GAMMA" §(cab | cba)
+            cab = "c" "a" §"b"
+            cba = "c" "b" §"a"
+        GAMMA_RE = /(?=GA\w+)/
+        ALPHA_RESUME = { !`BETA` !`GAMMA` /./ }
+        """
+
+    def test_several_resume_rules_innermost_rule_matching(self):
+        gr = grammar_provider(self.lang, fail_when=WARNING)()
+        content = 'ALPHA abc BETA bad GAMMA cab .'
+        cst = gr(content)
+        assert cst.error_flag
+        assert cst.content == content
+        assert cst.pick('alpha').content.startswith('ALPHA')
+        # because of resuming, there should be only on error message
+        assert len(cst.errors_sorted) == 1
+
+        content = 'ALPHA acb BETA bad GAMMA cab .'
+        cst = gr(content)
+        assert cst.error_flag
+        assert cst.content == content
+        assert cst.pick('alpha').content.startswith('ALPHA')
+        # because of resuming, there should be only on error message
+        assert len(cst.errors_sorted) == 2
+
+        content = 'ALPHA acb GAMMA cab .'
+        cst = gr(content)
+        assert cst.error_flag
+        assert cst.content == content
+        assert cst.pick('alpha').content.startswith('ALPHA')
+        # because of resuming, there should be only one error message
+        assert len(cst.errors_sorted) == 1, str(cst.errors_sorted)
+
 
 class TestInSeriesResume:
     def setup(self):
@@ -942,7 +1036,7 @@ class TestInterleaveResume:
         st = gr('AB_CDEFG.')
         # mandatory continuation error kicks in only, if the parsers before
         # the §-sign have been exhausted!
-        assert len(st.errors) == 2 and st.errors_sorted[1].code == MANDATORY_CONTINUATION
+        assert len(st.errors) == 2 and st.errors_sorted[0].code == MANDATORY_CONTINUATION
         st = gr('EXY EXYZ.')
         assert len(st.errors) == 1
 
@@ -962,8 +1056,6 @@ class TestInterleaveResume:
         assert not st.errors
 
         lang = """document = ({ `A` }) ° `B`\n"""
-        # code, errors, ast = compile_ebnf(lang, 'InterleaveTest', True)
-        # print(ast.as_sxpr())
         gr = grammar_provider(lang)()
         mini_suite(gr)
         st = gr('AABAA')
@@ -986,9 +1078,16 @@ VARIABLE   ::= /[A-Za-z]/, ~;
 
 
 class TestAlternativeEBNFSyntax:
+    def setup(self):
+        self.save = get_config_value('syntax_variant')
+        set_config_value('syntax_variant', 'heuristic')
+
+    def teardown(self):
+        set_config_value('syntax_variant', self.save)
+
     def test_alt_syntax(self):
         code, errors, ast = compile_ebnf(ArithmeticEBNF, preserve_AST=True)
-        assert not ast.error_flag
+        assert not ast.error_flag, str(ast.errors)
         arithmetic_grammer = compile_python_object(
             DHPARSER_IMPORTS.format(dhparser_parentdir=DHPARSER_PARENTDIR) + code)
         arithmetic_parser = arithmetic_grammer()
@@ -997,16 +1096,23 @@ class TestAlternativeEBNFSyntax:
 
 
 class TestSyntaxExtensions:
-    def test_difference(self):
-        lang = """
-            doc = /[A-Z]/ - /[D-F]/
-        """
-        parser = create_parser(lang)
-        st = parser("A")
-        assert not st.errors and st.tag_name == "doc" and st.content == "A"
-        st = parser("E")
-        assert st.errors and any(e.code == PARSER_STOPPED_BEFORE_END for e in st.errors)
+    def setup(self):
+        self.save = get_config_value('syntax_variant')
+        set_config_value('syntax_variant', 'strict')
 
+    def teardown(self):
+        set_config_value('syntax_variant', self.save)
+
+    # def test_difference(self):
+    #     lang = """
+    #         doc = /[A-Z]/ - /[D-F]/
+    #     """
+    #     parser = create_parser(lang)
+    #     st = parser("A")
+    #     assert not st.errors and st.tag_name == "doc" and st.content == "A"
+    #     st = parser("E")
+    #     assert st.errors and any(e.code == PARSER_STOPPED_BEFORE_END for e in st.errors)
+    #
     def test_any_char(self):
         lang = 'doc = "A".'
         parser = create_parser(lang)
@@ -1028,22 +1134,20 @@ class TestSyntaxExtensions:
         assert not st.errors
 
     def test_simple_char_range(self):
-        set_config_value('syntax_variant', EBNF_ANY_SYNTAX_STRICT)
+        set_config_value('syntax_variant', 'strict')
         lang = "Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]"
-        # print(raw_compileEBNF(lang).result)
         parser = create_parser(lang)
         st = parser('賌')
         assert st.as_sxpr() == '(Char "賌")'
 
     def test_full_char_range(self):
-        set_config_value('syntax_variant', EBNF_ANY_SYNTAX_HEURISTICAL)
+        set_config_value('syntax_variant', 'heuristic')
         lang = """
             Identifier <- IdentStart IdentCont* Spacing
             IdentCont  <- IdentStart / [0-9] 
             IdentStart <- [a-zA-Z_]
             Spacing    <- (´ ´ / ´\t´ / ´\n´)*      
             """
-        # print(raw_compileEBNF(lang).result)
         parser = create_parser(lang)
         st = parser('marke_8')
         assert not st.errors
@@ -1051,10 +1155,17 @@ class TestSyntaxExtensions:
         assert not st.errors
         st = parser('3tvp ')
         assert st.errors
-        set_config_value('syntax_variant', EBNF_ANY_SYNTAX_STRICT)
+        set_config_value('syntax_variant', 'strict')
 
 
 class TestModeSetting:
+    def setup(self):
+        self.save = get_config_value('syntax_variant')
+        set_config_value('syntax_variant', 'heuristic')
+
+    def teardown(self):
+        set_config_value('syntax_variant', self.save)
+
     testdoc = r"""# hey, you
 
         doc = sequence | re | char | char_range | char_range2 | multiple1 | multiple2 | multiple3 | mutliple4
@@ -1074,29 +1185,36 @@ class TestModeSetting:
 
     def test_setmode_getmode(self):
         gr = get_ebnf_grammar()
-        gr.mode = EBNF_ANY_SYNTAX_STRICT
-        assert gr.mode == EBNF_ANY_SYNTAX_STRICT
-        gr.mode = EBNF_REGULAR_EXPRESSION_SYNTAX
-        assert gr.mode == EBNF_REGULAR_EXPRESSION_SYNTAX
-        gr.mode = EBNF_PARSING_EXPRESSION_GRAMMAR_SYNTAX
-        assert gr.mode == EBNF_PARSING_EXPRESSION_GRAMMAR_SYNTAX
-        gr.mode = EBNF_ANY_SYNTAX_HEURISTICAL
-        assert gr.mode == EBNF_ANY_SYNTAX_HEURISTICAL
+        gr.mode = 'strict'
+        assert gr.mode == 'strict'
+        gr.mode = 'regex-like'
+        assert gr.mode == 'regex-like'
+        gr.mode = 'peg-like'
+        assert gr.mode == 'peg-like'
+        gr.mode = 'heuristic'
+        assert gr.mode == 'heuristic'
 
-        gr.mode = EBNF_CLASSIC_SYNTAX
-        assert gr.mode == EBNF_ANY_SYNTAX_STRICT
+        gr.mode = 'classic'
+        assert gr.mode == 'strict'
 
     def test_heuristic_mode(self):
         gr = get_ebnf_grammar()
-        gr.mode = EBNF_ANY_SYNTAX_STRICT
+        gr.mode = 'strict'
         st = gr(self.testdoc)
         assert st.errors
-        gr.mode = EBNF_ANY_SYNTAX_HEURISTICAL
+        gr.mode = 'heuristic'
         st = gr(self.testdoc)
-        assert not st.errors
+        assert not st.errors, str(st.errors)
 
 
 class TestAlternativeReordering:
+    def setup(self):
+        self.save = get_config_value('syntax_variant')
+        set_config_value('syntax_variant', 'strict')
+
+    def teardown(self):
+        set_config_value('syntax_variant', self.save)
+
     def test_reordering_alternative_literals(self):
         lang = """
             TokenizedType  ::= 'ID'
@@ -1156,23 +1274,214 @@ class TestAlternativeReordering:
         assert any(e.code == BAD_ORDER_OF_ALTERNATIVES for e in errors)
 
 
-class TestDrop:
+class TestTreeOptimization:
     def test_drop_anonymous(self):
         lang = """
-            @ drop = B
             doc = A B C
             A = `A`
             B = `B`
             C = `C`
         """
-        # print(raw_compileEBNF(lang).result)
         parser = create_parser(lang)
         st = parser('ABC')
-        assert str(st) == "AC"
-        parser = create_parser('@ anonymous = B\n' + lang)
+        assert str(st) == "ABC"
+        parser = create_parser('@ disposable = B\n@ drop = B\n' + lang)
         st = parser('ABC')
         assert str(st) == "AC"
 
+    def test_tree_reduction(self):
+        lang = """@reduction = none
+           root = 'A' 'B' (important | 'D')
+           important = 'C' 
+        """
+        parser = create_parser(lang)
+        st = parser('ABC')
+        assert st.as_sxpr() == '(root (:Text "A") (:Text "B") (:Alternative (important "C")))'
+        parser = create_parser(lang.replace('none', 'flatten'))
+        assert parser('ABC').as_sxpr() == '(root (:Text "A") (:Text "B") (important "C"))'
+        parser = create_parser(lang.replace('none', 'merge_treetops'))
+        assert parser('ABC').as_sxpr() == '(root (:Text "A") (:Text "B") (important "C"))'
+        assert parser('ABD').as_sxpr() == '(root "ABD")'
+        parser = create_parser(lang.replace('none', 'merge_leaves'))
+        assert parser('ABC').as_sxpr() == '(root (:Text "AB") (important "C"))'
+        assert parser('ABD').as_sxpr() == '(root "ABD")'
+        parser = create_parser(lang.replace('none', 'merge'))
+        assert parser('ABC').as_sxpr() == '(root (:Text "AB") (important "C"))'
+        assert parser('ABD').as_sxpr() == '(root "ABD")'
+
+
+class TestHeuristics:
+    def setup(self):
+        self.save = get_config_value('syntax_variant')
+        set_config_value('syntax_variant', 'fixed')
+
+    def teardown(self):
+        set_config_value('syntax_variant', self.save)
+
+    def test_heuristics(self):
+        lang = """
+            array = "[" [_element { "," _element }] "]"
+            _element = /[0-9]/
+        """
+        parser = create_parser(lang)
+        s = str(parser['array'])
+        assert s == "array = `[` [_element {`,` _element}] `]`", s
+
+        set_config_value('syntax_variant', 'strict')
+        parser = create_parser(lang + ' ')
+        s = str(parser['array'])
+        assert s == "array = `[` [_element {`,` _element}] `]`", s
+
+        # set_config_value('syntax_variant', 'heuristic')
+        # parser = create_parser(lang + '  ')  # + ' ' cheats the cache
+        # s = str(parser['array'])
+        # assert s != "array = `[` [_element {`,` _element}] `]`"
+
+
+class TestRuleOrder:
+    """Reordering of rules in order to minimize the number of Forward-Declarations."""
+
+    def test_rule_specification_order_does_not_matter(self):
+        normal_order = """
+            document = A
+            A = B
+            B = C
+            C = "Hallo Welt"
+            """
+        # set_config_value('compiled_EBNF_log', 'auto_reorder_test.log')
+        parser = create_parser(normal_order)
+        assert parser.B.__class__.__name__ != "Forward"
+
+        # Now B, should not be a Forward-Parser
+        reverse_order = """
+            document = A
+            C = "Hallo Welt"
+            B = C
+            A = B
+            """
+        parser = create_parser(reverse_order)
+        # If order of rule specification did not matter,
+        # B should not be a Forward-parser:
+        assert parser.B.__class__.__name__ != "Forward"
+
+    def test_recursive_root(self):
+        reverse_order = """
+            C = "Hallo Welt" [A]
+            B = C
+            A = B
+            """
+        # set_config_value('compiled_EBNF_log', 'auto_reorder_test.log')
+        parser = create_parser(reverse_order)
+        assert parser.B.__class__.__name__ != "Forward"
+
+    def test_recursive_root_detached_root(self):
+        reverse_order = """
+            C = "Hallo Welt"
+            B = C
+            A = B
+            """
+        # set_config_value('compiled_EBNF_log', 'auto_reorder_test.log')
+        parser = create_parser(reverse_order)
+        assert parser.B.__class__.__name__ != "Forward"
+
+    def test_resusive_paths(self):
+        lang = """
+            A = B | C
+            B = C [`+` A]
+            C = `x` [`-` A] [`*` B]
+        """
+        ebnf_grammar = get_ebnf_grammar()
+        st = ebnf_grammar(lang)
+        assert not st.errors, str(st.errors)
+        ebnf_transformer = get_ebnf_transformer()
+        ebnf_transformer(st)
+        ebnf_compiler = get_ebnf_compiler()
+        python_src = ebnf_compiler(st)
+        pathsA = ebnf_compiler.recursive_paths('A')
+        assert pathsA == frozenset({('A', 'B', 'C'), ('A', 'B'),
+                                    ('A', 'C', 'B'), ('A', 'C')})
+        pathsB = ebnf_compiler.recursive_paths('B')
+        assert pathsB == frozenset({('B', 'C', 'A'), ('B', 'A', 'C'),
+                                    ('B', 'A'), ('B', 'C')})
+        pathsC = ebnf_compiler.recursive_paths('C')
+        assert pathsC == frozenset({('C', 'A', 'B'), ('C', 'A'),
+                                    ('C', 'B'), ('C', 'B', 'A')})
+
+    def test_order_of_declarations(self):
+        """If the order of definitions changes, the following invariants should hold:
+        1. The number of Forward declarations is the same
+        2. No two Forwardly-declared symbols yield exactly the same set of recursive paths
+        3. The union of sets of recursive paths from all farwad-declared symbol
+           remains the same.
+        4. But: Depending on the order of definitions the symbols which are declared as
+           Forward symbols can change within the limits of the previous 3 invariantes.
+        """
+
+        def compileEBNF(lang):
+            ebnf_grammar = get_ebnf_grammar()
+            st = ebnf_grammar(lang)
+            assert not st.errors, str(st.errors)
+            ebnf_transformer = get_ebnf_transformer()
+            ebnf_transformer(st)
+            ebnf_compiler = get_ebnf_compiler()
+            ebnf_compiler(st)
+            return ebnf_compiler
+
+        def all_paths(compiler_obj):
+            paths = [normalize_circular_path(compiler_obj.recursive_paths(sym))
+                     for sym in compiler_obj.forward]
+            for i in range(len(paths)):
+                for k in range(i + 1, len(paths)):
+                    assert paths[i] ^ paths[k]
+            return paths[0].union(*paths[1:])
+
+        lang = """doc = A
+            A = B | C
+            B = C [`+` A]
+            C = `x` [`-` A] [`*` B]
+        """
+        compiler_obj = compileEBNF(lang)
+        A = all_paths(compiler_obj)
+        lang = """doc = A
+            A = B | C
+            C = `x` [`-` A] [`*` B]
+            B = C [`+` A]
+        """
+        compiler_obj = compileEBNF(lang)
+        B = all_paths(compiler_obj)
+        assert A == B, str(A) + str(B)
+
+        lang = """doc = A
+            A = B | C
+            B = C [`+` A]
+            C = `x` [`-` A]
+        """
+        compiler_obj = compileEBNF(lang)
+        A = all_paths(compiler_obj)
+        lang = """doc = A
+            A = B | C
+            C = `x` [`-` A]
+            B = C [`+` A]
+        """
+        compiler_obj = compileEBNF(lang)
+        B = all_paths(compiler_obj)
+        assert A == B, str(A) + str(B)
+
+        lang = """doc = A
+            A = B | C
+            B = C [`+` A]
+            C = `x` [`*` B]
+        """
+        compiler_obj = compileEBNF(lang)
+        A = all_paths(compiler_obj)
+        lang = """doc = A
+            A = B | C
+            C = `x` [`*` B]
+            B = C [`+` A]
+        """
+        compiler_obj = compileEBNF(lang)
+        B = all_paths(compiler_obj)
+        assert A == B, str(A) + str(B)
 
 if __name__ == "__main__":
     from DHParser.testing import runner
