@@ -30,18 +30,19 @@ sys.path.append(os.path.abspath(os.path.join(scriptpath, '..')))
 
 from DHParser.configuration import get_config_value, set_config_value
 from DHParser.toolkit import compile_python_object, re
-from DHParser.log import is_logging, log_ST, log_parsing_history
-from DHParser.error import Error, is_error, adjust_error_locations, MANDATORY_CONTINUATION, \
+from DHParser.log import is_logging, log_ST, log_parsing_history, start_logging
+from DHParser.error import Error, is_error, add_source_locations, MANDATORY_CONTINUATION, \
     MALFORMED_ERROR_STRING, MANDATORY_CONTINUATION_AT_EOF, RESUME_NOTICE, PARSER_STOPPED_BEFORE_END, \
-    PARSER_NEVER_TOUCHES_DOCUMENT
+    PARSER_NEVER_TOUCHES_DOCUMENT, CAPTURE_DROPPED_CONTENT_WARNING
 from DHParser.parse import ParserError, Parser, Grammar, Forward, TKN, ZeroOrMore, RE, \
     RegExp, Lookbehind, NegativeLookahead, OneOrMore, Series, Alternative, \
-    Interleave, UnknownParserError, CombinedParser, Text, EMPTY_NODE, Capture, Drop, Whitespace, \
-    GrammarError, Counted, Always, INFINITE
+    Interleave, CombinedParser, Text, EMPTY_NODE, Capture, Drop, Whitespace, \
+    GrammarError, Counted, Always, INFINITE, longest_match
 from DHParser import compile_source
 from DHParser.ebnf import get_ebnf_grammar, get_ebnf_transformer, get_ebnf_compiler, \
     parse_ebnf, DHPARSER_IMPORTS, compile_ebnf
-from DHParser.dsl import grammar_provider, create_parser, raw_compileEBNF
+from DHParser.dsl import grammar_provider, create_parser
+from DHParser.preprocess import gen_neutral_srcmap_func
 from DHParser.syntaxtree import Node, parse_sxpr
 from DHParser.stringview import StringView
 from DHParser.trace import set_tracer, trace_history, resume_notices_on
@@ -59,7 +60,8 @@ class TestWhitespace:
 
 class TestParserError:
     def test_parser_error_str(self):
-        pe = ParserError(Node('TAG', 'test').with_pos(0), StringView('Beispiel'), None, True)
+        pe = ParserError(Parser(), Node('TAG', 'test').with_pos(0),
+                         StringView('Beispiel'), None, first_throw=True)
         assert str(pe).find('Beispiel') >= 0 and str(pe).find('TAG') >= 0
 
     def test_false_lookahead_only_message(self):
@@ -106,19 +108,35 @@ class TestParserClass:
             root__ = word
         gr = MyGrammar()
         regex = gr['word'].parsers[-1].parser
-        result = gr.associated_symbol(regex).symbol
+        result = gr.associated_symbol__(regex).symbol
         assert result == 'word', result
 
 
 class TestInfiLoopsAndRecursion:
+    def setup(self):
+        pass
+        # set_config_value('history_tracking', True)
+        # set_config_value('resume_notices', True)
+        # start_logging('LOGS')
+
     def test_very_simple(self):
         minilang = """
             term = term (`*`|`/`) factor | factor
             factor = /[0-9]+/
             """
-        parser = grammar_provider(minilang)()
-        snippet = "5*4*3"
+        grammar_factory = grammar_provider(minilang)
+        parser = grammar_factory()
+        snippet = "5*4*3*2"
+        # set_tracer(parser, trace_history)
         st = parser(snippet)
+        if is_logging():
+            log_ST(st, 'test_LeftRecursion_very_simple.cst')
+            log_parsing_history(parser, 'test_LeftRecursion_very_simple')
+        assert not is_error(st.error_flag), str(st.errors)
+        st = parser("1*2*3*4*5*6*7*8*9")
+        # if is_logging():
+        #     log_ST(st, 'test_LeftRecursion_very_simple_2.cst')
+        #     log_parsing_history(parser, 'test_LeftRecursion_very_simple_2')
         assert not is_error(st.error_flag)
 
     def test_direct_left_recursion1(self):
@@ -132,17 +150,18 @@ class TestInfiLoopsAndRecursion:
         # print(raw_compileEBNF(minilang).result)
         assert parser
         syntax_tree = parser(snippet)
+        if is_logging():
+            log_ST(syntax_tree, "test_LeftRecursion_direct1.cst")
+            log_parsing_history(parser, "test_LeftRecursion_direct1")
         assert not is_error(syntax_tree.error_flag), str(syntax_tree.errors_sorted)
         assert snippet == syntax_tree.content, str(syntax_tree)
-        if is_logging():
-            log_ST(syntax_tree, "test_LeftRecursion_direct.cst")
-            log_parsing_history(parser, "test_LeftRecursion_direct")
 
     def test_direct_left_recursion2(self):
         minilang = """@literalws = right
             expr = ex
             ex   = expr ("+"|"-") term | term
-            term = term ("*"|"/") factor | factor
+            term = tr
+            tr   = term ("*"|"/") factor | factor
             factor = /[0-9]+/~
             """
         snippet = "9 + 8 + 7 + 6 + 5 + 3 * 4"
@@ -151,6 +170,9 @@ class TestInfiLoopsAndRecursion:
         syntax_tree = parser(snippet)
         assert not is_error(syntax_tree.error_flag), syntax_tree.errors_sorted
         assert snippet == syntax_tree.content
+        if is_logging():
+            log_ST(syntax_tree, "test_LeftRecursion_direct2.cst")
+            log_parsing_history(parser, "test_LeftRecursion_direct2")
 
     def test_indirect_left_recursion1(self):
         minilang = """@literalws = right
@@ -166,71 +188,103 @@ class TestInfiLoopsAndRecursion:
         assert not is_error(syntax_tree.error_flag), syntax_tree.errors_sorted
         snippet = "7 + 8 * 4"
         syntax_tree = parser(snippet)
+        if is_logging():
+            log_ST(syntax_tree, "test_LeftRecursion_indirect1.cst")
+            log_parsing_history(parser, "test_LeftRecursion_indirect1")
         assert not is_error(syntax_tree.error_flag), syntax_tree.errors_sorted
         snippet = "9 + 8 * (4 + 3)"
         syntax_tree = parser(snippet)
         assert not is_error(syntax_tree.error_flag), syntax_tree.errors_sorted
         assert snippet == syntax_tree.content
+        snippet = "9 + 8 * (4 - 3 / (5 - 1))"
+        syntax_tree = parser(snippet)
+        assert not is_error(syntax_tree.error_flag), syntax_tree.errors_sorted
+        assert snippet == syntax_tree.content
         if is_logging():
-            log_ST(syntax_tree, "test_LeftRecursion_indirect.cst")
-            log_parsing_history(parser, "test_LeftRecursion_indirect")
+            log_ST(syntax_tree, "test_LeftRecursion_indirect1.cst")
+            log_parsing_history(parser, "test_LeftRecursion_indirect1")
+
+    # BEWARE: EXPERIMENTAL TEST can be long running
+    def test_indirect_left_recursion2(self):
+        arithmetic_syntax = r"""@literalws = right
+            expression     = addition | subtraction  # | term
+            addition       = (expression | term) "+" (expression | term)
+            subtraction    = (expression | term) "-" (expression | term)
+            term           = multiplication | division  # | factor
+            multiplication = (term | factor) "*" (term | factor)
+            division       = (term | factor) "/" (term | factor)
+            factor         = [SIGN] ( NUMBER | VARIABLE | group ) { VARIABLE | group }
+            group          = "(" expression ")"
+            SIGN           = /[+-]/
+            NUMBER         = /(?:0|(?:[1-9]\d*))(?:\.\d+)?/~
+            VARIABLE       = /[A-Za-z]/~
+            """
+        arithmetic = grammar_provider(arithmetic_syntax)()
+        assert arithmetic
+        syntax_tree = arithmetic("(a + b) * (a - b)")
+        assert syntax_tree.errors
+        if is_logging():
+            log_ST(syntax_tree, "test_LeftRecursion_indirect2.cst")
+            log_parsing_history(arithmetic, "test_LeftRecursion_indirect2")
+
+    def test_indirect_left_recursion3(self):
+        arithmetic_syntax = r"""@literalws = right
+            expression     = addition | subtraction | term
+            addition       = (expression | term) "+" (expression | term)
+            subtraction    = (expression | term) "-" (expression | term)
+            term           = multiplication | division | factor
+            multiplication = (term | factor) "*" (term | factor)
+            division       = (term | factor) "/" (term | factor)
+            factor         = [SIGN] ( NUMBER | VARIABLE | group ) { VARIABLE | group }
+            group          = "(" expression ")"
+            SIGN           = /[+-]/
+            NUMBER         = /(?:0|(?:[1-9]\d*))(?:\.\d+)?/~
+            VARIABLE       = /[A-Za-z]/~
+            """
+        arithmetic = grammar_provider(arithmetic_syntax)()
+        assert arithmetic
+        syntax_tree = arithmetic("(a + b) * (a - b)")
+        assert not syntax_tree.errors
+        if is_logging():
+            log_ST(syntax_tree, "test_LeftRecursion_indirect3.cst")
+            log_parsing_history(arithmetic, "test_LeftRecursion_indirect3")
 
-    # # BEWARE: EXPERIMENTAL TEST can be long running
-    # def test_indirect_left_recursion2(self):
-    #     arithmetic_syntax = """@literalws = right
-    #         expression     = addition | subtraction
-    #         addition       = (expression | term) "+" (expression | term)
-    #         subtraction    = (expression | term) "-" (expression | term)
-    #         term           = multiplication | division
-    #         multiplication = (term | factor) "*" (term | factor)
-    #         division       = (term | factor) "/" (term | factor)
-    #         factor         = [SIGN] ( NUMBER | VARIABLE | group ) { VARIABLE | group }
-    #         group          = "(" expression ")"
-    #         SIGN           = /[+-]/
-    #         NUMBER         = /(?:0|(?:[1-9]\d*))(?:\.\d+)?/~
-    #         VARIABLE       = /[A-Za-z]/~
-    #         """
-    #     arithmetic = grammar_provider(arithmetic_syntax)()
-    #     arithmetic.left_recursion_depth__ = 2
-    #     assert arithmetic
-    #     syntax_tree = arithmetic("(a + b) * (a - b)")
-    #     assert syntax_tree.errors
 
     def test_break_inifnite_loop_ZeroOrMore(self):
         forever = ZeroOrMore(RegExp(''))
         result = Grammar(forever)('')  # infinite loops will automatically be broken
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
 
     def test_break_inifnite_loop_OneOrMore(self):
         forever = OneOrMore(RegExp(''))
         result = Grammar(forever)('')  # infinite loops will automatically be broken
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
 
     def test_break_infinite_loop_Counted(self):
         forever = Counted(Always(), (0, INFINITE))
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
         forever = Counted(Always(), (5, INFINITE))
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
         forever = Counted(Always(), (INFINITE, INFINITE))
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
         forever = Counted(Always(), (1000, INFINITE - 1))
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
 
     def test_break_infinite_loop_Interleave(self):
-        forever = Interleave(Always(), repetitions = [(0, INFINITE)])
+        forever = Interleave(Always(), repetitions=[(0, INFINITE)])
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
         forever = Interleave(Always(), Always(),
-                             repetitions = [(5, INFINITE), (INFINITE, INFINITE)])
+                             repetitions=[(5, INFINITE), (INFINITE, INFINITE)])
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
-        forever = Interleave(Always(), repetitions = [(1000, INFINITE - 1)])
+        assert repr(result) == "Node('root', '')", repr(result)
+        forever = Interleave(Always(), repetitions=[(1000, INFINITE - 1)])
         result = Grammar(forever)('')  # if this takes very long, something is wrong
-        assert repr(result) == "Node(':EMPTY', '')", repr(result)
+        assert repr(result) == "Node('root', '')", repr(result)
 
     # def test_infinite_loops(self):
     #     minilang = """forever = { // } \n"""
@@ -250,11 +304,11 @@ class TestInfiLoopsAndRecursion:
     #     snippet = " "
     #     syntax_tree = parser(snippet)
     #     assert any(e.code == INFINITE_LOOP for e in syntax_tree.errors)
-    #     res = parser.static_analysis()
+    #     res = parser.static_analysis__()
     #     assert res and res[0][2].code == INFINITE_LOOP
     #     minilang = """not_forever = { / / } \n"""
     #     parser = grammar_provider(minilang)()
-    #     res = parser.static_analysis()
+    #     res = parser.static_analysis__()
     #     assert not res
     #     set_config_value('static_analysis', save)
 
@@ -498,6 +552,7 @@ class TestSeries:
         st = parser("A_CD");  assert not st.error_flag
         st = parser("AB_D");  assert st.error_flag
         assert st.errors_sorted[0].code == MANDATORY_CONTINUATION
+        assert str(st.errors_sorted[0]).find("series") >= 0
         # transitivity of mandatory-operator
         st = parser("ABC_");  assert st.error_flag
         assert st.errors_sorted[0].code == MANDATORY_CONTINUATION
@@ -540,7 +595,7 @@ class TestSeries:
     def test_ebnf_serialization(self):
         ebnf_grammar = get_ebnf_grammar()
         # TODO: Add test here
-        ebnf = ebnf_grammar.as_ebnf()
+        ebnf = ebnf_grammar.as_ebnf__()
         # print(ebnf)
 
 
@@ -635,6 +690,45 @@ class TestErrorRecovery:
         assert len(st.errors) == 2 and any(err.code == RESUME_NOTICE for err in st.errors)
         assert 'Skipping' in str(st.errors_sorted[1])
 
+    def test_series_skip2(self):
+        grammar = r"""
+        @whitespace = vertical
+        @literalws = right
+        @document_skip = /\s+|(?=$)/
+        document = { sentence § sentence_continuation }+ EOF
+            sentence_continuation = &(sentence | EOF)
+        @sentence_skip = /\s+|(?=\.|$)/
+        sentence = { word § word_continuation }+ "."
+            word_continuation = &(word | `.` | EOF)
+        word = /[A-Za-z]+/~
+        EOF = !/./ 
+        """
+        data = "Time is out of joint. Oh cursed spite that I was ever born to set it right."
+        parser = grammar_provider(grammar)()
+        st = parser(data)
+        assert not st.errors, str(st.errors)
+        data2 = data.replace('cursed', 'cur?ed')
+        st = parser(data2)
+        assert len(st.errors) == 1
+        zombie = st.pick('ZOMBIE__')
+        assert zombie.content == '?ed ', zombie.content
+
+    def test_irrelevance_of_error_definition_order(self):
+        lang = """
+        document = series | /.*/
+        series = "A" "B" §"C" "D"
+        @series_skip = /(?=[A-Z])/
+        @series_error = '', "Unerwartetes Zeichen"
+        """
+        parser = grammar_provider(lang)()
+        st = parser('AB_D')
+        assert len(st.errors) == 1  # no additional "stopped before end"-error!
+        assert st.errors[0].message == "Unerwartetes Zeichen"
+        resume_notices_on(parser)
+        st = parser('AB_D')
+        assert len(st.errors) == 2 and any(err.code == RESUME_NOTICE for err in st.errors)
+        assert 'Skipping' in str(st.errors_sorted[1])
+
     def test_Interleave_skip(self):
         lang = """
         document = allof | /.*/
@@ -666,7 +760,7 @@ class TestPopRetrieve:
     mini_language = r"""
         document       = { text | codeblock }
         codeblock      = delimiter { text | (!:delimiter delimiter_sign) } ::delimiter
-        delimiter      = delimiter_sign  # never use delimiter between capture and pop except for retrival!
+        delimiter      = delimiter_sign  # never use delimiter between capture and pop except for retrieval!
         delimiter_sign = /`+/
         text           = /[^`]+/
         """
@@ -713,14 +807,16 @@ class TestPopRetrieve:
     def test_capture_assertions(self):
         try:
             _ = Grammar(Capture(Drop(Whitespace(r'\s*'))))
-            assert False, "GrammarError expected!"
+            assert False, "GrammarError expected"
         except GrammarError as ge:
-            pass
+            assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
+                "Capture-dropped-content-Warning expected"
         try:
             _ = Grammar(Capture(Series(Text(' '), Drop(Whitespace(r'\s*')))))
-            assert False, "ValueError expected!"
-        except GrammarError:
-            pass
+            assert False, "GrammarError expected"
+        except GrammarError as ge:
+            assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
+                "Capture-dropped-content-Warning expected"
         cp = Capture(RegExp(r'\w+'))
         cp.pname = "capture"
         _ = Grammar(cp)
@@ -753,6 +849,10 @@ class TestPopRetrieve:
         assert not syntax_tree.error_flag, str(syntax_tree.errors_sorted)
 
     def test_optional_match(self):
+        # from DHParser.dsl import compileEBNF
+        # src = compileEBNF(self.mini_lang4)
+        # print(src)
+        # return
         test1 = '<info>Hey, you</info>'
         st = self.minilang_parser4(test1)
         assert not st.error_flag, str(st.errors_sorted)
@@ -760,7 +860,12 @@ class TestPopRetrieve:
         st = self.minilang_parser4(test1)
         assert not st.error_flag
         test2 = '<info>Hey, you</>'
+        # set_config_value('history_tracking', True)
+        # set_tracer(self.minilang_parser4, trace_history)
+        # start_logging('LOGS')
         st = self.minilang_parser4(test2)
+        # log_parsing_history(self.minilang_parser4, "optional_match")
+        # print(st.as_sxpr())
         assert not st.error_flag
         test3 = '<info>Hey, <emph>you</></>'
         st = self.minilang_parser4(test3)
@@ -779,7 +884,7 @@ class TestPopRetrieve:
         assert not self.minilang_parser4.variables__['name']
         assert st.error_flag
 
-    def test_cache_neutrality(self):
+    def test_cache_neutrality_1(self):
         """Test that packrat-caching does not interfere with the variable-
         changing parsers: Capture and Retrieve."""
         lang = r"""@literalws = right
@@ -796,6 +901,46 @@ class TestPopRetrieve:
         st = gr(case)
         assert not st.error_flag, str(st.errors_sorted)
 
+    def test_cache_neutrality_2(self):
+        lang = r'''document = variantA | variantB
+            variantA  = delimiter `X` ::delimiter `!` 
+            variantB  = `A` delimiter ::delimiter `!` 
+            delimiter = `A` | `X`
+        '''
+        gr = grammar_provider(lang)()
+        case = 'AXA!'
+        st = gr(case)
+        assert not st.errors
+        case = 'AXX!'
+        # set_config_value('history_tracking', True)
+        # start_logging('LOGS')
+        # set_tracer(gr, trace_history)
+        st = gr(case)
+        # log_parsing_history(gr, 'test_cache_neutrality_2')
+        # print(st.as_sxpr())
+        assert not st.errors
+        assert str(st) == "AXX!"
+
+    def test_cache_neutrality_3(selfself):
+        lang = r'''document = variantA | variantB
+            variantA  = delimiter `X` [`---`] check ::delimiter `!` 
+            variantB  = `A` delimiter [`---`] check ::delimiter `!`
+            check = :delimiter 
+            delimiter = `A` | `X`
+        '''
+        gr = grammar_provider(lang)()
+        case = 'AX---XX!'
+        # set_config_value('history_tracking', True)
+        # start_logging('LOGS')
+        # set_tracer(gr, trace_history)
+        st = gr(case)
+        # log_parsing_history(gr, 'test_cache_neutrality_3')
+        # print(st.as_sxpr())
+        assert not st.errors, str(errors)
+        case = 'AXXX!'
+        st = gr(case)
+        assert not st.errors
+
     def test_single_line(self):
         teststr = "Anfang ```code block `` <- keine Ende-Zeichen ! ``` Ende"
         syntax_tree = self.minilang_parser(teststr)
@@ -865,11 +1010,11 @@ class TestPopRetrieve:
             value      = /\d+/~
             EOF        = !/./ [ :?defsign ]   # eat up captured defsigns
         """
-        # code, _, _ = compile_ebnf(lang)
-        # print(code)
+        # print(raw_compileEBNF(lang).result)
+        # set_config_value('compiled_EBNF_log', 'mylog.txt')
         parser = grammar_provider(lang)()
         st = parser("X := 1")
-        assert not st.error_flag
+        assert not st.error_flag, str(st.errors)
         st1 = st
         st = parser("")
         assert not st.error_flag
@@ -905,7 +1050,7 @@ class TestPopRetrieve:
         assert not st.errors
         assert st.equals(st1)
         st = parser('')
-        assert "EOF expected" in str(st.errors)
+        assert "'EOF' expected" in str(st.errors), st.as_sxpr()
 
 
 class TestWhitespaceHandling:
@@ -984,7 +1129,7 @@ EBNF_with_Errors = r"""# Test code with errors. All places marked by a "$" shoul
 @ comment    = /#.*(?:\n|$)/
 @ whitespace = /\s*/
 @ literalws  = right
-@ anonymous  = pure_elem, EOF
+@ disposable = pure_elem, EOF
 @ drop       = whitespace, EOF
 
 
@@ -1165,7 +1310,9 @@ class TestReentryAfterError:
             block_B = "x" "y" "z"
             """
         proc = """
-def next_valid_letter(text, start):
+def next_valid_letter(text, start, end):
+    L = len(text)
+    end = min(L, max(L, end))
     while start < len(text):
         if str(text[start]) in 'abcxyz':
             return start, 0
@@ -1228,7 +1375,6 @@ def next_valid_letter(text, start):
         gr = copy.deepcopy(get_ebnf_grammar())
         resume_notices_on(gr)
         cst = gr(EBNF_with_Errors)
-        adjust_error_locations(cst.errors, EBNF_with_Errors)
         locations = []
         for error in cst.errors_sorted:
             locations.append((error.line, error.column))
@@ -1257,7 +1403,7 @@ class TestUnknownParserError:
         try:
             gr("", "NonExistantParser")
             assert False, "UnknownParserError expected!"
-        except UnknownParserError:
+        except AttributeError:
             pass
 
 
@@ -1293,12 +1439,10 @@ class TestMetaParser:
     mp = CombinedParser()
     mp.grammar = Grammar()  # override placeholder warning
     mp.pname = "named"
-    mp.anonymous = False
+    mp.disposable = False
     mp.tag_name = mp.pname
 
     def test_return_value(self):
-        save = get_config_value('flatten_tree')
-        set_config_value('flatten_tree', True)
         nd = self.mp._return_value(Node('tagged', 'non-empty'))
         assert nd.tag_name == 'named', nd.as_sxpr()
         assert len(nd.children) == 1
@@ -1320,7 +1464,7 @@ class TestMetaParser:
         nd = self.mp._return_value(EMPTY_NODE)
         assert nd.tag_name == 'named' and not nd.children, nd.as_sxpr()
         self.mp.pname = ''
-        self.mp.anonymous = True
+        self.mp.disposable = True
         self.mp.tag_name = ':unnamed'
         nd = self.mp._return_value(Node('tagged', 'non-empty'))
         assert nd.tag_name == 'tagged', nd.as_sxpr()
@@ -1340,12 +1484,12 @@ class TestMetaParser:
         assert not nd.content
         assert self.mp._return_value(None) == EMPTY_NODE
         assert self.mp._return_value(EMPTY_NODE) == EMPTY_NODE
-        set_config_value('flatten_tree', save)
 
     def test_return_values(self):
         self.mp.pname = "named"
         self.mp.tag_name = self.mp.pname
         rv = self.mp._return_values((Node('tag', 'content'), EMPTY_NODE))
+        # print(rv.as_sxpr())
         assert rv[-1].tag_name != EMPTY_NODE.tag_name, rv[-1].tag_name
 
     def test_in_context(self):
@@ -1422,15 +1566,16 @@ class TestStaticAnalysis:
     def test_cannot_capture_dropped_content(self):
         p = Capture(Drop(Whitespace(" ")))
         try:
-            gr = Grammar(p)
+            _ = Grammar(p)
             assert False, "GrammarError expected"
-        except GrammarError:
-            pass
+        except GrammarError as ge:
+            assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
+                "Capture-dropped-content-Warning expected"
 
     def test_cyclical_ebnf_error(self):
         doc = Text('proper');  doc.pname = "doc"
         grammar = Grammar(doc)
-        # grammar.static_analysis()
+        # grammar.static_analysis__()
         lang = "doc = 'proper'  # this works!"
         lang1 = "doc = { doc }  # this parser never reaches a leaf parser."
         lang2 = """doc = word | sentence  # a more convoluted example
@@ -1446,6 +1591,71 @@ class TestStaticAnalysis:
         #     print(e)
 
 
+class TestMemoization:
+    def test_memoization(self):
+        words = r'''@literalws = right
+        list = word { ',' word } §EOF
+        word = wordA | wordB | wordC
+        wordA = `"` /[Aa]\w+/ '"'
+        wordB = `"` /[Bb]\w+/ '"'
+        wordC = `"` /[Cc]\w+/ '"'
+        EOF = /$/'''
+        grammar = create_parser(words, 'words')
+
+        # print(grammar.python_src__)
+        p1 = grammar.wordC.parsers[0]
+        p2 = grammar.wordB.parsers[0]
+        p3 = grammar.wordA.parsers[0]
+
+        ps1 = grammar.wordC.parsers[-1]
+        ps2 = grammar.wordB.parsers[-1]
+        ps3 = grammar.wordA.parsers[-1]
+
+        p4 = ps1.parsers[0]
+        p5 = ps1.parsers[0]
+        p6 = ps1.parsers[0]
+
+        assert (p.eq_class == p1.eq_class for p in (p2, p3, p4, p5, p6))
+        assert (p.eq_class == ps1.eq_class for p in (ps2, ps3))
+
+        cst = grammar('"camma", "beta", "alpha"')
+
+
+class TestStringAlternative:
+    def test_longest_match(self):
+        l = ['a', 'ab', 'ca', 'cd']
+        assert longest_match(l, '0abdd') == ''
+        assert longest_match(l, 'axcde') == 'a'
+        assert longest_match(l, 'cab') == 'ca'
+        assert longest_match(l, 'b') == ''
+        assert longest_match(l, 'x') == ''
+        assert longest_match(l, 'a') == 'a'
+        assert longest_match(l, 'ab') == 'ab'
+        assert longest_match(l, 'ca') == 'ca'
+        assert longest_match(l, 'cd') == 'cd'
+        assert longest_match(l, 'cb') == ''
+        assert longest_match(l, 'cdc') == 'cd'
+        assert longest_match(l, 'c') == ''
+        l = ['a', 'ab', 'abc', 'abcd']
+        assert longest_match(l, 'abxyz') == 'ab'
+        assert longest_match(l, 'abcdxyz') == 'abcd'
+        assert longest_match(l, 'abcxyz') == 'abc'
+        assert longest_match(l, 'axyz') == 'a'
+        assert longest_match(l, 'axyz') == 'a'
+        assert longest_match(l, 'ax') == 'a'
+        assert longest_match(l, '') == ''
+        assert longest_match([], 'a') == ''
+        assert longest_match([], '') == ''
+        l = ['abc', 'xy']
+        assert longest_match(l, 'xyzt', 2) == 'xy'
+        assert longest_match(l, 'abcdefg', 1) == 'abc'
+        assert longest_match(l, 'abcdefg', 2) == 'abc'
+        assert longest_match(l, 'ax12345', 2) == ''
+        assert longest_match(l, 'a', 2) == ''
+
+
+
 if __name__ == "__main__":
     from DHParser.testing import runner
     runner("", globals())
+
